/* trsfmt -- format a TRS-80 Model 3/4 disk */
/* 1998 A.R.Duell and placed under the GPL */

#include<stdio.h>
#include<string.h>
#include<unistd.h>
#include<fcntl.h>
#include<linux/fd.h>
#include<linux/fdreg.h>
#include<sys/ioctl.h>

#define SECTORS_PER_TRACK 18
#define GPL 0x0c

/* Floppy controller data rate selection codes */
#define RATE250 2
#define RATE300 1

int data_rate(int device)
  {
    struct floppy_drive_params params;
    int rate;

    ioctl(device,FDGETDRVPRM,&params);
    rate=(params.rps==6)?RATE300:RATE250;
    printf("rps=%d,rate=%d\n",params.rps,rate);
    return (rate);
  } 
  
void recalibrate(int device, int rate)
  {
    struct floppy_raw_cmd cmd;

    cmd.data=NULL;
    cmd.length=0;
    cmd.rate=rate;
    cmd.flags=FD_RAW_INTR;
    cmd.cmd[0]=FD_RECALIBRATE;
    cmd.cmd[1]=0;
    cmd.cmd_count=2;
    if(ioctl(device,FDRAWCMD,&cmd)<0)
      {
         fprintf(stderr,"Error on recalibrate\n");
         exit(1);
      }
  }

void seek(int device, int rate, int track, int dblstep)
  {
    struct floppy_raw_cmd cmd;

    cmd.data=NULL;
    cmd.length=0;
    cmd.rate=rate;
    cmd.flags=FD_RAW_INTR;
    cmd.cmd[0]=FD_SEEK;
    cmd.cmd[1]=0;
    cmd.cmd[2]=track*(1<<dblstep);
    cmd.cmd_count=3;
    if(ioctl(device,FDRAWCMD,&cmd)<0)
      {
         fprintf(stderr,"Error on seek to track %d\n",track);
         exit(1);
      }
  }

void format_track(int device, int rate, int track, int side)
  {
    unsigned char data[4*SECTORS_PER_TRACK];
    int skew_table[SECTORS_PER_TRACK]=
         {0,6,12,1,7,13,2,8,14,3,9,15,4,10,16,5,11,17};
    int i,j; /* counters */
    struct floppy_raw_cmd cmd;

    /* Make sector header table */
    for(i=0,j=0;i<SECTORS_PER_TRACK;i++)
      {
        data[j++]=track; /* C */
        data[j++]=side; /* H */
        data[j++]=skew_table[i]; /* R */
        data[j++]=1; /* N */
      }

    cmd.data=data;
    cmd.length=4*SECTORS_PER_TRACK;
    cmd.rate=rate;
    cmd.flags=FD_RAW_INTR | FD_RAW_WRITE;
    cmd.cmd[0]=FD_FORMAT;
    cmd.cmd[1]=side?4:0;
    cmd.cmd[2]=1; /* N */
    cmd.cmd[3]=SECTORS_PER_TRACK;
    cmd.cmd[4]=GPL;
    cmd.cmd[5]=0x6b;
    cmd.cmd_count=6;
    if (ioctl(device,FDRAWCMD,&cmd)<0)
      {
        fprintf(stderr,"Error formatting track %d\n",track);
        exit(1);
      }
  }

void format_disk(int device, int rate, int dblstep, int track_count, 
                 int side_count)
  {
    int track;
    int side;

    recalibrate(device,rate);
    for(track=0;track<track_count;track++)
      {
        seek(device,rate,track,dblstep);
        for(side=0;side<side_count;side++)
          {
            printf("Formatting track %d, side %d\n",track,side);
            format_track(device,rate,track,side);
          }
      }
  }

void usage(void)
  {
    printf("usage : trsfmt (-d) (-t tracks) (-s sides) device\n\n");
    printf("        -d - double step floppy drive\n");
    printf("        -t - set # tracks (default = 40)\n");
    printf("        -s - set # sides  (default = 1)\n\n");
    printf("This program is released under the GPL v2.0. See 'copying' for\n");
    printf("further details\n");
    exit(1);
  }

int main(int argc, char **argv)
  {
    int option; /* option character */
    int side_count=1; /* No of sides */
    int track_count=40; /* No of tracks */
    int dblstep=0; /* double step the floppy drive */
    int rate; /* data rate code */
    int device; /* file number of floppy device */

    optind=1;
    while ((option=getopt(argc,argv,"ds:t:?"))!=-1)
      {
        switch(option)
          {
            case 'd' : dblstep=1;
                       break;
            case 's' : side_count=atoi(optarg);
                       break;
            case 't' : track_count=atoi(optarg);
                       break;
            case '?' : usage();
          }
      }
    if (optind != argc-1) {usage();} /* Error if device name not given */
    if ((device=open(argv[optind],O_RDWR,0))==-1)
      {
        fprintf(stderr,"Error opening device %s\n",argv[optind]);
        exit(1);
      }
    rate=data_rate(device);
    format_disk(device,rate,dblstep,track_count,side_count);
    close(device);
    exit(0);
  }

